@@ -1,16 +1,12 @@
 /*
  * ZIP support routines for PhysicsFS.
  *
- * Please see the file LICENSE in the source's root directory.
+ * Please see the file LICENSE.txt in the source's root directory.
  *
  *  This file written by Ryan C. Gordon, with some peeking at "unzip.c"
  *   by Gilles Vollant.
  */
 
-#if HAVE_CONFIG_H
-#  include <config.h>
-#endif
-
 #if (defined PHYSFS_SUPPORTS_ZIP)
 
 #include <stdio.h>
@@ -27,8 +23,8 @@
 #include "physfs_internal.h"
 
 /*
- * A buffer of ZIP_READBUFSIZE is malloc() for each compressed file opened,
- *  and is free()'d when you close the file; compressed data is read into
+ * A buffer of ZIP_READBUFSIZE is allocated for each compressed file opened,
+ *  and is freed when you close the file; compressed data is read into
  *  this buffer, and then is decompressed into the buffer passed to
  *  PHYSFS_read().
  *
@@ -116,70 +112,33 @@ typedef struct
 #define UNIX_FILETYPE_SYMLINK 0120000
 
 
-static PHYSFS_sint64 ZIP_read(FileHandle *handle, void *buffer,
-                              PHYSFS_uint32 objSize, PHYSFS_uint32 objCount);
-static PHYSFS_sint64 ZIP_write(FileHandle *handle, const void *buffer,
-                               PHYSFS_uint32 objSize, PHYSFS_uint32 objCount);
-static int ZIP_eof(FileHandle *handle);
-static PHYSFS_sint64 ZIP_tell(FileHandle *handle);
-static int ZIP_seek(FileHandle *handle, PHYSFS_uint64 offset);
-static PHYSFS_sint64 ZIP_fileLength(FileHandle *handle);
-static int ZIP_fileClose(FileHandle *handle);
-static int ZIP_isArchive(const char *filename, int forWriting);
-static DirHandle *ZIP_openArchive(const char *name, int forWriting);
-static LinkedStringList *ZIP_enumerateFiles(DirHandle *h,
-                                            const char *dirname,
-                                            int omitSymLinks);
-static int ZIP_exists(DirHandle *h, const char *name);
-static int ZIP_isDirectory(DirHandle *h, const char *name, int *fileExists);
-static int ZIP_isSymLink(DirHandle *h, const char *name, int *fileExists);
-static PHYSFS_sint64 ZIP_getLastModTime(DirHandle *h, const char *n, int *e);
-static FileHandle *ZIP_openRead(DirHandle *h, const char *filename, int *e);
-static FileHandle *ZIP_openWrite(DirHandle *h, const char *filename);
-static FileHandle *ZIP_openAppend(DirHandle *h, const char *filename);
-static void ZIP_dirClose(DirHandle *h);
-static int zip_resolve(void *in, ZIPinfo *info, ZIPentry *entry);
-static int ZIP_remove(DirHandle *h, const char *name);
-static int ZIP_mkdir(DirHandle *h, const char *name);
-
-
-const PHYSFS_ArchiveInfo __PHYSFS_ArchiveInfo_ZIP =
+/*
+ * Bridge physfs allocation functions to zlib's format...
+ */
+static voidpf zlibPhysfsAlloc(voidpf opaque, uInt items, uInt size)
 {
-    "ZIP",
-    ZIP_ARCHIVE_DESCRIPTION,
-    "Ryan C. Gordon <icculus@clutteredmind.org>",
-    "http://icculus.org/physfs/",
-};
+    return(((PHYSFS_Allocator *) opaque)->Malloc(items * size));
+} /* zlibPhysfsAlloc */
 
-static const FileFunctions __PHYSFS_FileFunctions_ZIP =
+/*
+ * Bridge physfs allocation functions to zlib's format...
+ */
+static void zlibPhysfsFree(voidpf opaque, voidpf address)
 {
-    ZIP_read,       /* read() method       */
-    ZIP_write,      /* write() method      */
-    ZIP_eof,        /* eof() method        */
-    ZIP_tell,       /* tell() method       */
-    ZIP_seek,       /* seek() method       */
-    ZIP_fileLength, /* fileLength() method */
-    ZIP_fileClose   /* fileClose() method  */
-};
+    ((PHYSFS_Allocator *) opaque)->Free(address);
+} /* zlibPhysfsFree */
 
 
-const DirFunctions __PHYSFS_DirFunctions_ZIP =
+/*
+ * Construct a new z_stream to a sane state.
+ */
+static void initializeZStream(z_stream *pstr)
 {
-    &__PHYSFS_ArchiveInfo_ZIP,
-    ZIP_isArchive,          /* isArchive() method      */
-    ZIP_openArchive,        /* openArchive() method    */
-    ZIP_enumerateFiles,     /* enumerateFiles() method */
-    ZIP_exists,             /* exists() method         */
-    ZIP_isDirectory,        /* isDirectory() method    */
-    ZIP_isSymLink,          /* isSymLink() method      */
-    ZIP_getLastModTime,     /* getLastModTime() method */
-    ZIP_openRead,           /* openRead() method       */
-    ZIP_openWrite,          /* openWrite() method      */
-    ZIP_openAppend,         /* openAppend() method     */
-    ZIP_remove,             /* remove() method         */
-    ZIP_mkdir,              /* mkdir() method          */
-    ZIP_dirClose            /* dirClose() method       */
-};
+    memset(pstr, '\0', sizeof (z_stream));
+    pstr->zalloc = zlibPhysfsAlloc;
+    pstr->zfree = zlibPhysfsFree;
+    pstr->opaque = &allocator;
+} /* initializeZStream */
 
 
 static const char *zlib_error_string(int rc)
@@ -191,12 +150,12 @@ static const char *zlib_error_string(int rc)
 #ifndef _WIN32_WCE
         case Z_ERRNO: return(strerror(errno));
 #endif
-        case Z_NEED_DICT: return(ERR_ZLIB_NEED_DICT);
-        case Z_DATA_ERROR: return(ERR_ZLIB_DATA_ERROR);
-        case Z_MEM_ERROR: return(ERR_ZLIB_MEMORY_ERROR);
-        case Z_BUF_ERROR: return(ERR_ZLIB_BUFFER_ERROR);
-        case Z_VERSION_ERROR: return(ERR_ZLIB_VERSION_ERROR);
-        default: return(ERR_ZLIB_UNKNOWN_ERROR);
+        case Z_NEED_DICT: return(ERR_NEED_DICT);
+        case Z_DATA_ERROR: return(ERR_DATA_ERROR);
+        case Z_MEM_ERROR: return(ERR_MEMORY_ERROR);
+        case Z_BUF_ERROR: return(ERR_BUFFER_ERROR);
+        case Z_VERSION_ERROR: return(ERR_VERSION_ERROR);
+        default: return(ERR_UNKNOWN_ERROR);
     } /* switch */
 
     return(NULL);
@@ -239,10 +198,10 @@ static int readui16(void *in, PHYSFS_uint16 *val)
 } /* readui16 */
 
 
-static PHYSFS_sint64 ZIP_read(FileHandle *handle, void *buf,
+static PHYSFS_sint64 ZIP_read(fvoid *opaque, void *buf,
                               PHYSFS_uint32 objSize, PHYSFS_uint32 objCount)
 {
-    ZIPfileinfo *finfo = (ZIPfileinfo *) (handle->opaque);
+    ZIPfileinfo *finfo = (ZIPfileinfo *) opaque;
     ZIPentry *entry = finfo->entry;
     PHYSFS_sint64 retval = 0;
     PHYSFS_sint64 maxread = ((PHYSFS_sint64) objSize) * objCount;
@@ -313,29 +272,29 @@ static PHYSFS_sint64 ZIP_read(FileHandle *handle, void *buf,
 } /* ZIP_read */
 
 
-static PHYSFS_sint64 ZIP_write(FileHandle *handle, const void *buf,
+static PHYSFS_sint64 ZIP_write(fvoid *opaque, const void *buf,
                                PHYSFS_uint32 objSize, PHYSFS_uint32 objCount)
 {
     BAIL_MACRO(ERR_NOT_SUPPORTED, -1);
 } /* ZIP_write */
 
 
-static int ZIP_eof(FileHandle *handle)
+static int ZIP_eof(fvoid *opaque)
 {
-    ZIPfileinfo *finfo = ((ZIPfileinfo *) (handle->opaque));
+    ZIPfileinfo *finfo = (ZIPfileinfo *) opaque;
     return(finfo->uncompressed_position >= finfo->entry->uncompressed_size);
 } /* ZIP_eof */
 
 
-static PHYSFS_sint64 ZIP_tell(FileHandle *handle)
+static PHYSFS_sint64 ZIP_tell(fvoid *opaque)
 {
-    return(((ZIPfileinfo *) (handle->opaque))->uncompressed_position);
+    return(((ZIPfileinfo *) opaque)->uncompressed_position);
 } /* ZIP_tell */
 
 
-static int ZIP_seek(FileHandle *handle, PHYSFS_uint64 offset)
+static int ZIP_seek(fvoid *opaque, PHYSFS_uint64 offset)
 {
-    ZIPfileinfo *finfo = (ZIPfileinfo *) (handle->opaque);
+    ZIPfileinfo *finfo = (ZIPfileinfo *) opaque;
     ZIPentry *entry = finfo->entry;
     void *in = finfo->handle;
 
@@ -360,7 +319,7 @@ static int ZIP_seek(FileHandle *handle, PHYSFS_uint64 offset)
         {
             /* we do a copy so state is sane if inflateInit2() fails. */
             z_stream str;
-            memset(&str, '\0', sizeof (z_stream));
+            initializeZStream(&str);
             if (zlib_err(inflateInit2(&str, -MAX_WBITS)) != Z_OK)
                 return(0);
 
@@ -381,7 +340,7 @@ static int ZIP_seek(FileHandle *handle, PHYSFS_uint64 offset)
             if (maxread > sizeof (buf))
                 maxread = sizeof (buf);
 
-            if (ZIP_read(handle, buf, maxread, 1) != 1)
+            if (ZIP_read(finfo, buf, maxread, 1) != 1)
                 return(0);
         } /* while */
     } /* else */
@@ -390,26 +349,25 @@ static int ZIP_seek(FileHandle *handle, PHYSFS_uint64 offset)
 } /* ZIP_seek */
 
 
-static PHYSFS_sint64 ZIP_fileLength(FileHandle *handle)
+static PHYSFS_sint64 ZIP_fileLength(fvoid *opaque)
 {
-    ZIPfileinfo *finfo = (ZIPfileinfo *) (handle->opaque);
+    ZIPfileinfo *finfo = (ZIPfileinfo *) opaque;
     return(finfo->entry->uncompressed_size);
 } /* ZIP_fileLength */
 
 
-static int ZIP_fileClose(FileHandle *handle)
+static int ZIP_fileClose(fvoid *opaque)
 {
-    ZIPfileinfo *finfo = (ZIPfileinfo *) (handle->opaque);
+    ZIPfileinfo *finfo = (ZIPfileinfo *) opaque;
     BAIL_IF_MACRO(!__PHYSFS_platformClose(finfo->handle), NULL, 0);
 
     if (finfo->entry->compression_method != COMPMETH_NONE)
         inflateEnd(&finfo->stream);
 
     if (finfo->buffer != NULL)
-        free(finfo->buffer);
+        allocator.Free(finfo->buffer);
 
-    free(finfo);
-    free(handle);
+    allocator.Free(finfo);
     return(1);
 } /* ZIP_fileClose */
 
@@ -426,7 +384,7 @@ static PHYSFS_sint64 zip_find_end_of_central_dir(void *in, PHYSFS_sint64 *len)
     PHYSFS_uint32 extra = 0;
 
     filelen = __PHYSFS_platformFileLength(in);
-    BAIL_IF_MACRO(filelen == -1, NULL, 0);
+    BAIL_IF_MACRO(filelen == -1, NULL, 0);  /* !!! FIXME: unlocalized string */
     BAIL_IF_MACRO(filelen > 0xFFFFFFFF, "ZIP bigger than 2 gigs?!", 0);
 
     /*
@@ -489,6 +447,8 @@ static PHYSFS_sint64 zip_find_end_of_central_dir(void *in, PHYSFS_sint64 *len)
             break;
 
         filepos -= (maxread - 4);
+        if (filepos < 0)
+            filepos = 0;
     } /* while */
 
     BAIL_IF_MACRO(!found, ERR_NOT_AN_ARCHIVE, -1);
@@ -539,10 +499,10 @@ static void zip_free_entries(ZIPentry *entries, PHYSFS_uint32 max)
     {
         ZIPentry *entry = &entries[i];
         if (entry->name != NULL)
-            free(entry->name);
+            allocator.Free(entry->name);
     } /* for */
 
-    free(entries);
+    allocator.Free(entries);
 } /* zip_free_entries */
 
 
@@ -669,6 +629,8 @@ static void zip_expand_symlink_path(char *path)
     } /* while */
 } /* zip_expand_symlink_path */
 
+/* (forward reference: zip_follow_symlink and zip_resolve call each other.) */
+static int zip_resolve(void *in, ZIPinfo *info, ZIPentry *entry);
 
 /*
  * Look for the entry named by (path). If it exists, resolve it, and return
@@ -694,7 +656,7 @@ static ZIPentry *zip_follow_symlink(void *in, ZIPinfo *info, char *path)
         } /* else */
     } /* if */
 
-    free(path);
+    allocator.Free(path);
     return(entry);
 } /* zip_follow_symlink */
 
@@ -713,7 +675,7 @@ static int zip_resolve_symlink(void *in, ZIPinfo *info, ZIPentry *entry)
 
     BAIL_IF_MACRO(!__PHYSFS_platformSeek(in, entry->offset), NULL, 0);
 
-    path = (char *) malloc(size + 1);
+    path = (char *) allocator.Malloc(size + 1);
     BAIL_IF_MACRO(path == NULL, ERR_OUT_OF_MEMORY, 0);
     
     if (entry->compression_method == COMPMETH_NONE)
@@ -722,15 +684,15 @@ static int zip_resolve_symlink(void *in, ZIPinfo *info, ZIPentry *entry)
     else  /* symlink target path is compressed... */
     {
         z_stream stream;
-        PHYSFS_uint32 compsize = entry->compressed_size;
-        PHYSFS_uint8 *compressed = (PHYSFS_uint8 *) malloc(compsize);
+        PHYSFS_uint32 complen = entry->compressed_size;
+        PHYSFS_uint8 *compressed = (PHYSFS_uint8*) __PHYSFS_smallAlloc(complen);
         if (compressed != NULL)
         {
-            if (__PHYSFS_platformRead(in, compressed, compsize, 1) == 1)
+            if (__PHYSFS_platformRead(in, compressed, complen, 1) == 1)
             {
-                memset(&stream, '\0', sizeof (z_stream));
+                initializeZStream(&stream);
                 stream.next_in = compressed;
-                stream.avail_in = compsize;
+                stream.avail_in = complen;
                 stream.next_out = (unsigned char *) path;
                 stream.avail_out = size;
                 if (zlib_err(inflateInit2(&stream, -MAX_WBITS)) == Z_OK)
@@ -742,12 +704,12 @@ static int zip_resolve_symlink(void *in, ZIPinfo *info, ZIPentry *entry)
                     rc = ((rc == Z_OK) || (rc == Z_STREAM_END));
                 } /* if */
             } /* if */
-            free(compressed);
+            __PHYSFS_smallFree(compressed);
         } /* if */
     } /* else */
 
     if (!rc)
-        free(path);
+        allocator.Free(path);
     else
     {
         path[entry->uncompressed_size] = '\0';    /* null-terminate it. */
@@ -876,7 +838,7 @@ static int zip_version_does_symlinks(PHYSFS_uint32 version)
 } /* zip_version_does_symlinks */
 
 
-static int zip_entry_is_symlink(ZIPentry *entry)
+static int zip_entry_is_symlink(const ZIPentry *entry)
 {
     return((entry->resolved == ZIP_UNRESOLVED_SYMLINK) ||
            (entry->resolved == ZIP_BROKEN_SYMLINK) ||
@@ -964,7 +926,7 @@ static int zip_load_entry(void *in, ZIPentry *entry, PHYSFS_uint32 ofs_fixup)
     entry->resolved = (zip_has_symlink_attr(entry, external_attr)) ?
                             ZIP_UNRESOLVED_SYMLINK : ZIP_UNRESOLVED_FILE;
 
-    entry->name = (char *) malloc(fnamelen + 1);
+    entry->name = (char *) allocator.Malloc(fnamelen + 1);
     BAIL_IF_MACRO(entry->name == NULL, ERR_OUT_OF_MEMORY, 0);
     if (__PHYSFS_platformRead(in, entry->name, fnamelen, 1) != 1)
         goto zip_load_entry_puked;
@@ -983,39 +945,46 @@ static int zip_load_entry(void *in, ZIPentry *entry, PHYSFS_uint32 ofs_fixup)
     return(1);  /* success. */
 
 zip_load_entry_puked:
-    free(entry->name);
+    allocator.Free(entry->name);
     return(0);  /* failure. */
 } /* zip_load_entry */
 
 
 static int zip_entry_cmp(void *_a, PHYSFS_uint32 one, PHYSFS_uint32 two)
 {
-    ZIPentry *a = (ZIPentry *) _a;
-    return(strcmp(a[one].name, a[two].name));
+    if (one != two)
+    {
+        const ZIPentry *a = (const ZIPentry *) _a;
+        return(strcmp(a[one].name, a[two].name));
+    } /* if */
+
+    return 0;
 } /* zip_entry_cmp */
 
 
 static void zip_entry_swap(void *_a, PHYSFS_uint32 one, PHYSFS_uint32 two)
 {
-    ZIPentry tmp;
-    ZIPentry *first = &(((ZIPentry *) _a)[one]);
-    ZIPentry *second = &(((ZIPentry *) _a)[two]);
-    memcpy(&tmp, first, sizeof (ZIPentry));
-    memcpy(first, second, sizeof (ZIPentry));
-    memcpy(second, &tmp, sizeof (ZIPentry));
+    if (one != two)
+    {
+        ZIPentry tmp;
+        ZIPentry *first = &(((ZIPentry *) _a)[one]);
+        ZIPentry *second = &(((ZIPentry *) _a)[two]);
+        memcpy(&tmp, first, sizeof (ZIPentry));
+        memcpy(first, second, sizeof (ZIPentry));
+        memcpy(second, &tmp, sizeof (ZIPentry));
+    } /* if */
 } /* zip_entry_swap */
 
 
-static int zip_load_entries(void *in, DirHandle *dirh,
+static int zip_load_entries(void *in, ZIPinfo *info,
                             PHYSFS_uint32 data_ofs, PHYSFS_uint32 central_ofs)
 {
-    ZIPinfo *info = (ZIPinfo *) dirh->opaque;
     PHYSFS_uint32 max = info->entryCount;
     PHYSFS_uint32 i;
 
     BAIL_IF_MACRO(!__PHYSFS_platformSeek(in, central_ofs), NULL, 0);
 
-    info->entries = (ZIPentry *) malloc(sizeof (ZIPentry) * max);
+    info->entries = (ZIPentry *) allocator.Malloc(sizeof (ZIPentry) * max);
     BAIL_IF_MACRO(info->entries == NULL, ERR_OUT_OF_MEMORY, 0);
 
     for (i = 0; i < max; i++)
@@ -1032,11 +1001,10 @@ static int zip_load_entries(void *in, DirHandle *dirh,
 } /* zip_load_entries */
 
 
-static int zip_parse_end_of_central_dir(void *in, DirHandle *dirh,
+static int zip_parse_end_of_central_dir(void *in, ZIPinfo *info,
                                         PHYSFS_uint32 *data_start,
                                         PHYSFS_uint32 *central_dir_ofs)
 {
-    ZIPinfo *zipinfo = (ZIPinfo *) dirh->opaque;
     PHYSFS_uint32 ui32;
     PHYSFS_uint16 ui16;
     PHYSFS_sint64 len;
@@ -1063,8 +1031,8 @@ static int zip_parse_end_of_central_dir(void *in, DirHandle *dirh,
     BAIL_IF_MACRO(!readui16(in, &ui16), NULL, 0);
 
     /* total number of entries in the central dir */
-    BAIL_IF_MACRO(!readui16(in, &zipinfo->entryCount), NULL, 0);
-    BAIL_IF_MACRO(ui16 != zipinfo->entryCount, ERR_UNSUPPORTED_ARCHIVE, 0);
+    BAIL_IF_MACRO(!readui16(in, &info->entryCount), NULL, 0);
+    BAIL_IF_MACRO(ui16 != info->entryCount, ERR_UNSUPPORTED_ARCHIVE, 0);
 
     /* size of the central directory */
     BAIL_IF_MACRO(!readui32(in, &ui32), NULL, 0);
@@ -1100,86 +1068,62 @@ static int zip_parse_end_of_central_dir(void *in, DirHandle *dirh,
 } /* zip_parse_end_of_central_dir */
 
 
-static DirHandle *zip_allocate_dirhandle(const char *name)
+static ZIPinfo *zip_create_zipinfo(const char *name)
 {
     char *ptr;
-    ZIPinfo *info;
-    DirHandle *retval = malloc(sizeof (DirHandle));
-    BAIL_IF_MACRO(retval == NULL, ERR_OUT_OF_MEMORY, NULL);
-
-    memset(retval, '\0', sizeof (DirHandle));
-
-    info = (ZIPinfo *) malloc(sizeof (ZIPinfo));
-    if (info == NULL)
-    {
-        free(retval);
-        BAIL_MACRO(ERR_OUT_OF_MEMORY, NULL);
-    } /* if */
-
+    ZIPinfo *info = (ZIPinfo *) allocator.Malloc(sizeof (ZIPinfo));
+    BAIL_IF_MACRO(info == NULL, ERR_OUT_OF_MEMORY, 0);
     memset(info, '\0', sizeof (ZIPinfo));
 
-    ptr = (char *) malloc(strlen(name) + 1);
+    ptr = (char *) allocator.Malloc(strlen(name) + 1);
     if (ptr == NULL)
     {
-        free(info);
-        free(retval);
+        allocator.Free(info);
         BAIL_MACRO(ERR_OUT_OF_MEMORY, NULL);
     } /* if */
 
     info->archiveName = ptr;
     strcpy(info->archiveName, name);
-    retval->opaque = info;
-    retval->funcs = &__PHYSFS_DirFunctions_ZIP;
-    return(retval);
-} /* zip_allocate_dirhandle */
+    return(info);
+} /* zip_create_zipinfo */
 
 
-static DirHandle *ZIP_openArchive(const char *name, int forWriting)
+static void *ZIP_openArchive(const char *name, int forWriting)
 {
-    DirHandle *retval = NULL;
     void *in = NULL;
+    ZIPinfo *info = NULL;
     PHYSFS_uint32 data_start;
     PHYSFS_uint32 cent_dir_ofs;
-    int success = 0;
 
     BAIL_IF_MACRO(forWriting, ERR_ARC_IS_READ_ONLY, NULL);
 
     if ((in = __PHYSFS_platformOpenRead(name)) == NULL)
-        goto zip_openarchive_end;
+        goto zip_openarchive_failed;
     
-    if ((retval = zip_allocate_dirhandle(name)) == NULL)
-        goto zip_openarchive_end;
+    if ((info = zip_create_zipinfo(name)) == NULL)
+        goto zip_openarchive_failed;
 
-    if (!zip_parse_end_of_central_dir(in, retval, &data_start, &cent_dir_ofs))
-        goto zip_openarchive_end;
+    if (!zip_parse_end_of_central_dir(in, info, &data_start, &cent_dir_ofs))
+        goto zip_openarchive_failed;
 
-    if (!zip_load_entries(in, retval, data_start, cent_dir_ofs))
-        goto zip_openarchive_end;
+    if (!zip_load_entries(in, info, data_start, cent_dir_ofs))
+        goto zip_openarchive_failed;
 
-    success = 1;  /* ...and we're good to go.  :)  */
+    __PHYSFS_platformClose(in);
+    return(info);
 
-zip_openarchive_end:
-    if (!success)  /* clean up for failures. */
+zip_openarchive_failed:
+    if (info != NULL)
     {
-        if (retval != NULL)
-        {
-            if (retval->opaque != NULL)
-            {
-                if (((ZIPinfo *) (retval->opaque))->archiveName != NULL)
-                    free(((ZIPinfo *) (retval->opaque))->archiveName);
-
-                free(retval->opaque);
-            } /* if */
-
-            free(retval);
-            retval = NULL;
-        } /* if */
+        if (info->archiveName != NULL)
+            allocator.Free(info->archiveName);
+        allocator.Free(info);
     } /* if */
 
     if (in != NULL)
-        __PHYSFS_platformClose(in);  /* Close this even with success. */
+        __PHYSFS_platformClose(in);
 
-    return(retval);
+    return(NULL);
 } /* ZIP_openArchive */
 
 
@@ -1208,9 +1152,9 @@ static PHYSFS_sint32 zip_find_start_of_dir(ZIPinfo *info, const char *path,
         if (rc == 0)
         {
             char ch = name[dlen];
-            if (ch < '/') /* make sure this isn't just a substr match. */
+            if ('/' < ch) /* make sure this isn't just a substr match. */
                 rc = -1;
-            else if (ch > '/')
+            else if ('/' > ch)
                 rc = 1;
             else 
             {
@@ -1236,19 +1180,37 @@ static PHYSFS_sint32 zip_find_start_of_dir(ZIPinfo *info, const char *path,
 } /* zip_find_start_of_dir */
 
 
-static LinkedStringList *ZIP_enumerateFiles(DirHandle *h,
-                                            const char *dirname,
-                                            int omitSymLinks)
+/*
+ * Moved to seperate function so we can use alloca then immediately throw
+ *  away the allocated stack space...
+ */
+static void doEnumCallback(PHYSFS_EnumFilesCallback cb, void *callbackdata,
+                           const char *odir, const char *str, PHYSFS_sint32 ln)
 {
-    ZIPinfo *info = ((ZIPinfo *) h->opaque);
-    LinkedStringList *retval = NULL, *p = NULL;
+    char *newstr = __PHYSFS_smallAlloc(ln + 1);
+    if (newstr == NULL)
+        return;
+
+    memcpy(newstr, str, ln);
+    newstr[ln] = '\0';
+    cb(callbackdata, odir, newstr);
+    __PHYSFS_smallFree(newstr);
+} /* doEnumCallback */
+
+
+static void ZIP_enumerateFiles(dvoid *opaque, const char *dname,
+                               int omitSymLinks, PHYSFS_EnumFilesCallback cb,
+                               const char *origdir, void *callbackdata)
+{
+    ZIPinfo *info = ((ZIPinfo *) opaque);
     PHYSFS_sint32 dlen, dlen_inc, max, i;
 
-    i = zip_find_start_of_dir(info, dirname, 0);
-    BAIL_IF_MACRO(i == -1, ERR_NO_SUCH_FILE, NULL);
+    i = zip_find_start_of_dir(info, dname, 0);
+    if (i == -1)  /* no such directory. */
+        return;
 
-    dlen = strlen(dirname);
-    if ((dlen > 0) && (dirname[dlen - 1] == '/')) /* ignore trailing slash. */
+    dlen = strlen(dname);
+    if ((dlen > 0) && (dname[dlen - 1] == '/')) /* ignore trailing slash. */
         dlen--;
 
     dlen_inc = ((dlen > 0) ? 1 : 0) + dlen;
@@ -1256,7 +1218,7 @@ static LinkedStringList *ZIP_enumerateFiles(DirHandle *h,
     while (i < max)
     {
         char *e = info->entries[i].name;
-        if ((dlen) && ((strncmp(e, dirname, dlen) != 0) || (e[dlen] != '/')))
+        if ((dlen) && ((strncmp(e, dname, dlen) != 0) || (e[dlen] != '/')))
             break;  /* past end of this dir; we're done. */
 
         if ((omitSymLinks) && (zip_entry_is_symlink(&info->entries[i])))
@@ -1266,7 +1228,7 @@ static LinkedStringList *ZIP_enumerateFiles(DirHandle *h,
             char *add = e + dlen_inc;
             char *ptr = strchr(add, '/');
             PHYSFS_sint32 ln = (PHYSFS_sint32) ((ptr) ? ptr-add : strlen(add));
-            retval = __PHYSFS_addToLinkedStringList(retval, &p, add, ln);
+            doEnumCallback(cb, callbackdata, origdir, add, ln);
             ln += dlen_inc;  /* point past entry to children... */
 
             /* increment counter and skip children of subdirs... */
@@ -1278,26 +1240,24 @@ static LinkedStringList *ZIP_enumerateFiles(DirHandle *h,
             } /* while */
         } /* else */
     } /* while */
-
-    return(retval);
 } /* ZIP_enumerateFiles */
 
 
-static int ZIP_exists(DirHandle *h, const char *name)
+static int ZIP_exists(dvoid *opaque, const char *name)
 {
     int isDir;    
-    ZIPinfo *info = (ZIPinfo *) h->opaque;
+    ZIPinfo *info = (ZIPinfo *) opaque;
     ZIPentry *entry = zip_find_entry(info, name, &isDir);
     return((entry != NULL) || (isDir));
 } /* ZIP_exists */
 
 
-static PHYSFS_sint64 ZIP_getLastModTime(DirHandle *h,
+static PHYSFS_sint64 ZIP_getLastModTime(dvoid *opaque,
                                         const char *name,
                                         int *fileExists)
 {
     int isDir;
-    ZIPinfo *info = (ZIPinfo *) h->opaque;
+    ZIPinfo *info = (ZIPinfo *) opaque;
     ZIPentry *entry = zip_find_entry(info, name, &isDir);
 
     *fileExists = ((isDir) || (entry != NULL));
@@ -1309,9 +1269,9 @@ static PHYSFS_sint64 ZIP_getLastModTime(DirHandle *h,
 } /* ZIP_getLastModTime */
 
 
-static int ZIP_isDirectory(DirHandle *h, const char *name, int *fileExists)
+static int ZIP_isDirectory(dvoid *opaque, const char *name, int *fileExists)
 {
-    ZIPinfo *info = (ZIPinfo *) h->opaque;
+    ZIPinfo *info = (ZIPinfo *) opaque;
     int isDir;
     ZIPentry *entry = zip_find_entry(info, name, &isDir);
 
@@ -1340,10 +1300,10 @@ static int ZIP_isDirectory(DirHandle *h, const char *name, int *fileExists)
 } /* ZIP_isDirectory */
 
 
-static int ZIP_isSymLink(DirHandle *h, const char *name, int *fileExists)
+static int ZIP_isSymLink(dvoid *opaque, const char *name, int *fileExists)
 {
     int isDir;
-    ZIPentry *entry = zip_find_entry((ZIPinfo *) h->opaque, name, &isDir);
+    const ZIPentry *entry = zip_find_entry((ZIPinfo *) opaque, name, &isDir);
     *fileExists = ((isDir) || (entry != NULL));
     BAIL_IF_MACRO(entry == NULL, NULL, 0);
     return(zip_entry_is_symlink(entry));
@@ -1374,11 +1334,10 @@ static void *zip_get_file_handle(const char *fn, ZIPinfo *inf, ZIPentry *entry)
 } /* zip_get_file_handle */
 
 
-static FileHandle *ZIP_openRead(DirHandle *h, const char *fnm, int *fileExists)
+static fvoid *ZIP_openRead(dvoid *opaque, const char *fnm, int *fileExists)
 {
-    ZIPinfo *info = (ZIPinfo *) h->opaque;
+    ZIPinfo *info = (ZIPinfo *) opaque;
     ZIPentry *entry = zip_find_entry(info, fnm, NULL);
-    FileHandle *retval = NULL;
     ZIPfileinfo *finfo = NULL;
     void *in;
 
@@ -1388,75 +1347,104 @@ static FileHandle *ZIP_openRead(DirHandle *h, const char *fnm, int *fileExists)
     in = zip_get_file_handle(info->archiveName, info, entry);
     BAIL_IF_MACRO(in == NULL, NULL, NULL);
 
-    if ( ((retval = (FileHandle *) malloc(sizeof (FileHandle))) == NULL) ||
-         ((finfo = (ZIPfileinfo *) malloc(sizeof (ZIPfileinfo))) == NULL) )
+    finfo = (ZIPfileinfo *) allocator.Malloc(sizeof (ZIPfileinfo));
+    if (finfo == NULL)
     {
-        if (retval)
-            free(retval);
         __PHYSFS_platformClose(in);
         BAIL_MACRO(ERR_OUT_OF_MEMORY, NULL);
     } /* if */
 
-    retval->opaque = (void *) finfo;
-    retval->funcs = &__PHYSFS_FileFunctions_ZIP;
-    retval->dirHandle = h;
-
     memset(finfo, '\0', sizeof (ZIPfileinfo));
     finfo->handle = in;
     finfo->entry = ((entry->symlink != NULL) ? entry->symlink : entry);
+    initializeZStream(&finfo->stream);
     if (finfo->entry->compression_method != COMPMETH_NONE)
     {
         if (zlib_err(inflateInit2(&finfo->stream, -MAX_WBITS)) != Z_OK)
         {
-            ZIP_fileClose(retval);
+            ZIP_fileClose(finfo);
             return(NULL);
         } /* if */
 
-        finfo->buffer = (PHYSFS_uint8 *) malloc(ZIP_READBUFSIZE);
+        finfo->buffer = (PHYSFS_uint8 *) allocator.Malloc(ZIP_READBUFSIZE);
         if (finfo->buffer == NULL)
         {
-            ZIP_fileClose(retval);
+            ZIP_fileClose(finfo);
             BAIL_MACRO(ERR_OUT_OF_MEMORY, NULL);
         } /* if */
     } /* if */
 
-    return(retval);
+    return(finfo);
 } /* ZIP_openRead */
 
 
-static FileHandle *ZIP_openWrite(DirHandle *h, const char *filename)
+static fvoid *ZIP_openWrite(dvoid *opaque, const char *filename)
 {
     BAIL_MACRO(ERR_NOT_SUPPORTED, NULL);
 } /* ZIP_openWrite */
 
 
-static FileHandle *ZIP_openAppend(DirHandle *h, const char *filename)
+static fvoid *ZIP_openAppend(dvoid *opaque, const char *filename)
 {
     BAIL_MACRO(ERR_NOT_SUPPORTED, NULL);
 } /* ZIP_openAppend */
 
 
-static void ZIP_dirClose(DirHandle *h)
+static void ZIP_dirClose(dvoid *opaque)
 {
-    ZIPinfo *zi = (ZIPinfo *) (h->opaque);
+    ZIPinfo *zi = (ZIPinfo *) (opaque);
     zip_free_entries(zi->entries, zi->entryCount);
-    free(zi->archiveName);
-    free(zi);
-    free(h);
+    allocator.Free(zi->archiveName);
+    allocator.Free(zi);
 } /* ZIP_dirClose */
 
 
-static int ZIP_remove(DirHandle *h, const char *name)
+static int ZIP_remove(dvoid *opaque, const char *name)
 {
     BAIL_MACRO(ERR_NOT_SUPPORTED, 0);
 } /* ZIP_remove */
 
 
-static int ZIP_mkdir(DirHandle *h, const char *name)
+static int ZIP_mkdir(dvoid *opaque, const char *name)
 {
     BAIL_MACRO(ERR_NOT_SUPPORTED, 0);
 } /* ZIP_mkdir */
 
+
+const PHYSFS_ArchiveInfo __PHYSFS_ArchiveInfo_ZIP =
+{
+    "ZIP",
+    ZIP_ARCHIVE_DESCRIPTION,
+    "Ryan C. Gordon <icculus@icculus.org>",
+    "http://icculus.org/physfs/",
+};
+
+
+const PHYSFS_Archiver __PHYSFS_Archiver_ZIP =
+{
+    &__PHYSFS_ArchiveInfo_ZIP,
+    ZIP_isArchive,          /* isArchive() method      */
+    ZIP_openArchive,        /* openArchive() method    */
+    ZIP_enumerateFiles,     /* enumerateFiles() method */
+    ZIP_exists,             /* exists() method         */
+    ZIP_isDirectory,        /* isDirectory() method    */
+    ZIP_isSymLink,          /* isSymLink() method      */
+    ZIP_getLastModTime,     /* getLastModTime() method */
+    ZIP_openRead,           /* openRead() method       */
+    ZIP_openWrite,          /* openWrite() method      */
+    ZIP_openAppend,         /* openAppend() method     */
+    ZIP_remove,             /* remove() method         */
+    ZIP_mkdir,              /* mkdir() method          */
+    ZIP_dirClose,           /* dirClose() method       */
+    ZIP_read,               /* read() method           */
+    ZIP_write,              /* write() method          */
+    ZIP_eof,                /* eof() method            */
+    ZIP_tell,               /* tell() method           */
+    ZIP_seek,               /* seek() method           */
+    ZIP_fileLength,         /* fileLength() method     */
+    ZIP_fileClose           /* fileClose() method      */
+};
+
 #endif  /* defined PHYSFS_SUPPORTS_ZIP */
 
 /* end of zip.c ... */
